
/********************************************
fin.c

libmawk changes (C) 2009-2012, Tibor 'Igor2' Palinkas;
based on mawk code coming with the below copyright:

copyright 1991, 1992.  Michael D. Brennan

This is a source file for mawk, an implementation of
the AWK programming language.

Mawk is distributed without warranty under the terms of
the GNU General Public License, version 2, 1991.
********************************************/

#ifndef _POSIX_SOURCE
/* needed on some systems to get strdup() */
#define _POSIX_SOURCE
#endif
#include <stdlib.h>
#include "mawk.h"
#include "fin.h"
#include "memory.h"
#include "bi_vars.h"
#include "field.h"
#include "symtype.h"
#include "scan.h"
#include "vio.h"
#include <string.h>

#warning TODO: this should be gone:
#include <stdio.h>

#ifndef	  NO_FCNTL_H
#include <fcntl.h>
#endif

/* This file handles input files.  Opening, closing,
   buffering and (most important) splitting files into
   records, mawk_FINgets().
*/

extern int isatty(int);

int mawk_is_cmdline_assign(mawk_state_t *, char *);	/* also used by init */
/* open a FIN* by filename.
   It's the main stream if main_flag is set.
   Recognizes "-" as stdin.
*/

FIN *mawk_FINopen(mawk_state_t *MAWK, char *filename, int main_flag)
{
	register FIN *fin;
	vio_t *vf;

	vf = mawk_vio_finopen(MAWK, filename, main_flag);

	if (vf == NULL)
		return NULL;

	fin = ZMALLOC(FIN);
	fin->vf = vf;
	fin->flags = main_flag ? (MAIN_FLAG | START_FLAG) : START_FLAG;
	fin->buffp = fin->buff = (char *) zmalloc(MAWK, BUFFSZ + 1);
	if (fin->buff == NULL)
		return NULL;
	fin->nbuffs = 1;
	fin->buff[0] = 0;
	fin->artificial_size = 0;
	fin->artificial_buff = NULL;

	return fin;
}

/* frees the buffer and fd, but leaves FIN structure until
   the user calls close() */
void mawk_FINsemi_close(mawk_state_t *MAWK, register FIN *fin)
{
#define dead ((void *)MAWK)
	if (fin->buff != dead) {
		zfree(MAWK, fin->buff, fin->nbuffs * BUFFSZ + 1);

		mawk_vio_close(MAWK, fin->vf);
		fin->buff = fin->buffp = dead;	/* marks it semi_closed */
	}
	/* else was already semi_closed */
#undef dead
}

/* user called close() on input file */
void mawk_FINclose(mawk_state_t *MAWK, FIN *fin)
{
	mawk_FINsemi_close(MAWK, fin);
	ZFREE(fin);
}

/*--------
  target is big enough to hold size + 1 chars
  on exit the back of the target is zero terminated
 *--------------*/
unsigned mawk_fillbuff(mawk_state_t *MAWK, FIN *fin, register char *target, unsigned size)
{
	register int r;
	unsigned entry_size = size;

	if ((fin != NULL) && (fin->artificial_size > 0)) {
		if (size < fin->artificial_size) {	/* read part of the input */
			memcpy(target, fin->artificial_buff, size);
			target += size;
			fin->artificial_size = fin->artificial_size - size;
			memmove(fin->artificial_buff, fin->artificial_buff + size, fin->artificial_size);
			fin->artificial_size -= size;
			size = 0;
		}
		else {											/* size >= artificial_size, we can read all input */
			memcpy(target, fin->artificial_buff, fin->artificial_size);
			target += fin->artificial_size;
			free(fin->artificial_buff);
			size -= fin->artificial_size;
			fin->artificial_buff = NULL;
			fin->artificial_size = 0;
		}
	}
	else {
		while (size)
			switch (r = mawk_vio_read(MAWK, fin->vf, target, size)) {
			case -2:
				return mawk_FIN_nomore;
			case -1:
				mawk_errmsg(MAWK, errno, "read error");
				mawk_exitval(MAWK, 2, -1);

			case 0:
				goto out;

			default:
				target += r;
				size -= r;
				break;
			}
	}
out:
	*target = 0;
	return entry_size - size;
}

/* convert file-descriptor to FIN*.
   It's the main stream if main_flag is set
*/
FIN *mawk_FIN_vio_open(mawk_state_t *MAWK, vio_t *vf, int main_flag)
{
	FIN *fin;

#warning TODO: this is copied from FINopen
	fin = ZMALLOC(FIN);
	fin->vf = vf;
	fin->flags = main_flag ? (MAIN_FLAG | START_FLAG) : START_FLAG;
	fin->buffp = fin->buff = (char *) zmalloc(MAWK, BUFFSZ + 1);
	if (fin->buff == NULL)
		return NULL;
	fin->nbuffs = 1;
	fin->buff[0] = 0;
	fin->artificial_size = 0;
	fin->artificial_buff = NULL;

	if (vf == NULL)
		return fin;

	mawk_vio_fixstdin(MAWK, vf);

	return fin;
}
