
/********************************************
types.h

libmawk changes (C) 2009-2010, Tibor 'Igor2' Palinkas;
based on mawk code coming with the below copyright:

copyright 1991, Michael D. Brennan

This is a source file for mawk, an implementation of
the AWK programming language.

Mawk is distributed without warranty under the terms of
the GNU General Public License, version 2, 1991.
********************************************/

#ifndef  MAWK_TYPES_H
#define  MAWK_TYPES_H

#include  <libmawk/sizes.h>
#include  <libmawk/num.h>

/*  CELL  types  */

#define  C_NOINIT                0
#define  C_NUM                   1
#define  C_STRING                2
#define  C_STRNUM                3
#define  C_MBSTRN                4
				/*could be STRNUM, has not been checked */
#define  C_RE                    5
#define  C_SPACE                 6
				/* split on space */
#define  C_SNULL                 7
				/* split on the empty string  */
#define  C_REPL                  8
				/* a replacement string   '\&' changed to &  */
#define  C_REPLV                 9
				/* a vector replacement -- broken on &  */
#define  NUM_CELL_TYPES         10

/* these defines are used to check types for two
   CELLs which are adjacent in memory */

#define  TWO_NOINITS  (2*(1<<C_NOINIT))
#define  TWO_NUMS     (2*(1<<C_NUM))
#define  TWO_STRINGS  (2*(1<<C_STRING))
#define  TWO_STRNUMS  (2*(1<<C_STRNUM))
#define  TWO_MBSTRNS  (2*(1<<C_MBSTRN))
#define  NOINIT_AND_NUM     ((1<<C_NOINIT)+(1<<C_NUM))
#define  NOINIT_AND_STRING  ((1<<C_NOINIT)+(1<<C_STRING))
#define  NOINIT_AND_STRNUM  ((1<<C_NOINIT)+(1<<C_STRNUM))
#define  NUM_AND_STRING     ((1<<C_NUM)+(1<<C_STRING))
#define  NUM_AND_STRNUM     ((1<<C_STRNUM)+(1<<C_NUM))
#define  STRING_AND_STRNUM  ((1<<C_STRING)+(1<<C_STRNUM))
#define  NOINIT_AND_MBSTRN  ((1<<C_NOINIT)+(1<<C_MBSTRN))
#define  NUM_AND_MBSTRN     ((1<<C_NUM)+(1<<C_MBSTRN))
#define  STRING_AND_MBSTRN  ((1<<C_STRING)+(1<<C_MBSTRN))
#define  STRNUM_AND_MBSTRN  ((1<<C_STRNUM)+(1<<C_MBSTRN))

typedef struct {
	unsigned len;
	unsigned short ref_cnt;
	char str[2];
} STRING;

/* number of bytes more than the characters to store a
   string */
#define  STRING_OH   (sizeof(STRING)-1)


typedef struct cell {
	short type;
	short vcnt;										/* only used if type == C_REPLV   */
	PTR ptr;
	mawk_num_t dval;
} CELL;


/* all builtins are passed the evaluation stack pointer and
   return its new value, here is the type */

#include "mawk.h"
typedef CELL *(*PF_CP) (mawk_state_t *, CELL *);

/* an element of code (instruction) */
typedef union {
	int op;
	PTR ptr;
} INST;

#endif /* MAWK_TYPES_H */
