/* virtual I/O */
typedef enum mawk_vio_open_mode_e {
	MAWK_VIO_O_TRUNC,
	MAWK_VIO_O_APPEND,
	MAWK_VIO_I
} mawk_vio_open_mode_t;


/* close and flush files */
int mawk_vio_close(mawk_state_t *MAWK, vio_t *vf);
int mawk_vio_flush(mawk_state_t *MAWK, vio_t *vf);


/* open and close pipe (background process) */
vio_t *mawk_vio_open_pipe(mawk_state_t *MAWK, char *name, int type);

/* open a normal file */
vio_t *mawk_vio_open(mawk_state_t *MAWK, const char *name, mawk_vio_open_mode_t mode);

/* called after some fin opens: if vf is stdin and mawk is running
   interactive, vio should do the necessary fixes on vf for the proper
   buffering; this is an stdio buffering hack */
mawk_vio_fixstdin(mawk_state_t *MAWK, vio_t *vf);

/* set up and register stdout and stderr */
void mawk_vio_setup_stdouts(mawk_state_t * MAWK);

/* return 0 if a vf is a standard output (stdout or stderr) */
int mawk_vio_is_stdouts(mawk_state_t *MAWK, vio_t *vf);

/* check the error state of a file node (ferror()-style) */
int mawk_vio_error(mawk_state_t *MAWK, vio_t *vf);


/* write data to a vio * */
int mawk_vio_putc(mawk_state_t *MAWK, vio_t *vf, char c);
int mawk_vio_write_str(mawk_state_t *MAWK, vio_t *vf, const char *str);
int mawk_vio_write(mawk_state_t *MAWK, vio_t *vf, void *data, int len);
int mawk_vio_printf(mawk_state_t *MAWK, vio_t *vf, const char *fmt, ...);

/* change buffering on a vf - happens mostly on stdout */
int mawk_vio_setbuf(mawk_state_t *MAWK, vio_t *vf, int buf_enable);

/* return -2 if there's no more input at the moment */
int mawk_vio_read(mawk_state_t *MAWK, vio_t *f, char *dst, long int size);
int mawk_vio_fgetsable(mawk_state_t *MAWK, vio_t *vf);
char *mawk_vio_fgets(mawk_state_t *MAWK, vio_t *vf, char *s, int size);


/* open "stdin" as FIN */
FIN *mawk_fin_fd0_open(mawk_state_t *MAWK, int main_flag);

/* open a filename for FIN (input); if it's "-", it should be stdin */
vio_t *mawk_vio_finopen(mawk_state_t *MAWK, char *filename, int main_flag);

